#include <io.h>
#include "bdp.h"

#define fileExist(filename) access(filename, 0) // F_OK = 0
#define ifOrFail(function, message) if(function) { \
	fputs(message, stderr); \
	fputs(function, stderr); \
	return 1; \
}

ICE_KEY *ik;

long clientVersion;
long pazCount;
long fileCount;
PazFile *pazFile;
FileBlock *fileBlock;
FolderName *folderName;
PathHash *pathHash;
char **filename;

char* fileCheck(){
	if (fileExist("Paz\\pad00000.meta") == -1) return "access denied.";
	mkdir("BDP_BACKUP");
	//TODO: backup file

	return 0;
}

void initKey(){
	ik = ice_key_create(0);
	ice_key_set(ik, (unsigned char *) MAGIC);
}

char* readAndDecrypt(FILE* inputFile, char* outputFilename){
	FILE *outputFile = fopen(outputFilename, "wb");
	unsigned char* encryptedText;
	long length = 0;
	fread(&length, 4, 1, inputFile);
	encryptedText = calloc(length, 1);
	long length2 = fread(encryptedText, 1, length, inputFile);
	if (length != length2) {
		return "dump size differ.";
	}

	long cuts = length / 8;
	while (cuts--){
		ice_key_decrypt(ik, encryptedText, encryptedText);
		fwrite(encryptedText, 1, 8, outputFile);
		encryptedText += 8;
	}
	free(encryptedText);
	fclose(outputFile);

	return 0;
}

char* parse(){
	FILE *meta = fopen("Paz/pad00000.meta", "rb");
	fread(&clientVersion, 4, 1, meta);
	fread(&pazCount, 4, 1, meta); // get paz file count

	pazFile = (PazFile*) calloc(pazCount, sizeof(PazFile));

	for (long i = 0; i < pazCount; i++){
		fread(&pazFile[i].number, 4, 1, meta);
		fread(&pazFile[i].hash, 4, 1, meta);
		fread(&pazFile[i].size, 4, 1, meta);
	}
	fread(&fileCount, 4, 1, meta); //get folder count

	// backup unknown section
	unsigned char *unknownSection = calloc(1, 256000);
	fread(unknownSection, 1, 256000, meta);
	FILE *unknownSectionDump = fopen("BDP_BACKUP\\unknown.dat", "wb");
	fwrite(unknownSection, 1, 256000, unknownSectionDump);
	fclose(unknownSectionDump);
	free(unknownSection);

	FILE *fileUnformatList;
	if(DEBUG){
		fileUnformatList = fopen("BDP_BACKUP\\fileUnformatList.csv", "w");
		fprintf(fileUnformatList,"hash,folderNum,fileNum,pazNum,offset,zsize,size\n");
	}


	long clearFileCount = fileCount - 8000;
	fileBlock = (FileBlock*) calloc(clearFileCount, sizeof(FileBlock));
	for (long i = 0; i < clearFileCount; i++){
	    fread(&fileBlock[i].hash, 4, 1, meta);
	    fread(&fileBlock[i].folderNum, 4, 1, meta);
	    fread(&fileBlock[i].fileNum, 4, 1, meta);
	    fread(&fileBlock[i].pazNum, 4, 1, meta);
	    fread(&fileBlock[i].offset, 4, 1, meta);
	    fread(&fileBlock[i].zsize, 4, 1, meta);
	    fread(&fileBlock[i].size, 4, 1, meta);
	    if(DEBUG){
			fprintf(fileUnformatList,"%ld,%ld,%ld,%ld,%ld,%ld,%ld\n", fileBlock[i].hash, fileBlock[i].folderNum, fileBlock[i].fileNum, fileBlock[i].pazNum, fileBlock[i].offset, fileBlock[i].zsize, fileBlock[i].size);
	    }
	}
	if(DEBUG){
		fclose(fileUnformatList);
	}

	readAndDecrypt(meta, "BDP_BACKUP\\foldername.dat");
	readAndDecrypt(meta, "BDP_BACKUP\\filename.dat");

	fclose(meta);

	return 0;
}

void pickupNames(){
	FILE *foldernameDump = fopen("BDP_BACKUP\\foldername.dat", "rb");
	FILE *filenameDump = fopen("BDP_BACKUP\\filename.dat", "rb");
	FILE *fileList, *folderList;
	char* buffer;
	long ptr, filenameNumber, folderNumber;

	if(DEBUG){
		fileList = fopen("BDP_BACKUP\\fileDump.csv", "w");
		folderList = fopen("BDP_BACKUP\\folderDump.csv", "w");
	}

	folderName = (FolderName*) calloc(fileCount, sizeof(FolderName));
	folderNumber = 0;
	while(!feof(foldernameDump)){
		ptr = 0;
		buffer = (char *) calloc(1,128);
		fread(&folderName[folderNumber].num, 4, 1, foldernameDump);
		fread(&folderName[folderNumber].sub, 4, 1, foldernameDump);
		do {
			fread(buffer+ptr, 1, 1, foldernameDump);
			ptr++;
		} while(*(buffer+ptr-1));
		if(*(buffer)){
			folderName[folderNumber].name = buffer;
			if(DEBUG){
				fprintf(folderList, "%ld,%ld,%s\n", folderName[folderNumber].num, folderName[folderNumber].sub, buffer);
			}
		}
		folderNumber++;
	}
	fclose(foldernameDump);

	filename = (char **) calloc(fileCount, 64);
	filenameNumber = 0;
	while(!feof(filenameDump)){
		ptr = 0;
		buffer = (char *) calloc(1,64);
		do {
			fread(buffer+ptr, 1, 1, filenameDump);
			ptr++;
		} while(*(buffer+ptr-1));
		if(*(buffer)){
			*(filename+filenameNumber) = buffer;
			if(DEBUG){
				fprintf(fileList, "%ld,%s\n", filenameNumber, buffer);
			}
		}
		filenameNumber++;
	}

	if(DEBUG){
		fclose(fileList);
		fclose(folderList);
	}
	fclose(filenameDump);
}

void restructIndex(){
	char *buffer = (char*) calloc(1, 256);
	long clearFileCount = fileCount - 8000;
	PathHash *s;
	pathHash = NULL;
	FILE *fileList;
	if(DEBUG){
		fileList = fopen("BDP_BACKUP\\fileList.csv", "w");
	}
	for (long i = 0; i < clearFileCount; i++){
		s = (PathHash*) malloc(sizeof(PathHash));
		s->path = (char*) calloc(1, 256);
		sprintf(buffer, "%s%s", folderName[fileBlock[i].folderNum].name, filename[fileBlock[i].fileNum]);
		strcpy(s->path, buffer);
		s->fileBlockNum = i;
		HASH_ADD_KEYPTR(hh, pathHash, s->path, strlen(s->path), s);

		if(DEBUG){
			fprintf(fileList, "pad%05ld.PAZ,%ld,%ld,%s,%ld\n",
						fileBlock[i].pazNum,
						fileBlock[i].folderNum,
						fileBlock[i].fileNum,
						s->path,
						fileBlock[i].hash
					);
		}
	}
}

int init(){

	ifOrFail(fileCheck(), "[ERROR] File system error: ");
	initKey();
	ifOrFail(parse(), "[ERROR] File parse error: ");
	pickupNames();
	restructIndex();

	return 0;
}
